import argparse
import json
import huggingface_hub
from transformers import AutoTokenizer  # type: ignore[import-untyped]
import llguidance


def main() -> None:
    parser = argparse.ArgumentParser(
        description="Command line interface for LL Guidance."
    )
    parser.add_argument(
        "--tokenizer", help="Tokenizer name", default="microsoft/Phi-3.5-mini-instruct"
    )
    parser.add_argument("--lark", help="Lark grammar file")
    parser.add_argument("--json-schema", help="JSON schema file")
    parser.add_argument("--text", help="File containing simulated generated text")
    parser.add_argument("--log-level", help="Log level", default=1, type=int)
    parser.add_argument("--ff-tokens", help="Enable fast-forward tokens", action="store_true")
    parser.add_argument("--backtrack", help="Enable backtracking", action="store_true")
    args = parser.parse_args()
    tokenizer: str = args.tokenizer

    tok_name = huggingface_hub.hf_hub_download(tokenizer, "tokenizer.json")
    with open(tok_name, "r") as f:
        text = f.read()
    tok = llguidance.LLTokenizer(text)

    grm = {}
    if args.lark:
        with open(args.lark, "r") as f:
            grm["lark_grammar"] = f.read()
    if args.json_schema:
        with open(args.json_schema, "r") as f:
            grm["json_schema"] = json.loads(f.read())

    if not grm:
        raise ValueError("No grammar provided; need --lark or --json-schema")

    tokens = []

    if args.text:
        with open(args.text, "r") as f:
            text = f.read()
        hf_tok = AutoTokenizer.from_pretrained(tokenizer)
        tokens = hf_tok.encode(text, add_special_tokens=False)

    interp = llguidance.LLInterpreter(
        tok,
        json.dumps({"grammars": [grm]}),
        enable_ff_tokens=args.ff_tokens,
        enable_backtrack=args.backtrack,
        log_level=args.log_level,
    )
    interp.start_without_prompt()

    print(tok.dbg_tokens(tokens))
    for t in tokens:
        mask, r = interp.compute_mask()
        obj = json.loads(r)
        for p in obj["progress"]:
            if p["object"] != "text":
                print("\n  ", end="")
                print(p)
        print(tok.dbg_tokens([t]), end=" ")
        assert mask
        if mask[t] == 0:
            print("Token not in mask")
            break
        interp.commit_token(t)
    print("\n")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

from dataclasses import dataclass
from enum import Enum, auto
from typing import Optional, List, Dict, Tuple, Iterator
from abc import ABC, abstractmethod
import re


@dataclass
class Position:
    """Tracks position in source text for error reporting"""

    text: str
    pos: int

    def advance(self, n: int = 1) -> "Position":
        return Position(self.text, self.pos + n)

    def current(self) -> str:
        return self.text[self.pos] if self.pos < len(self.text) else ""

    def peek(self, n: int = 1) -> str:
        return self.text[self.pos : self.pos + n]

    def __str__(self) -> str:
        line_no = self.text.count("\n", 0, self.pos) + 1
        pref = self.text[max(0, self.pos - 20) : self.pos]
        suff = self.text[self.pos : self.pos + 20]
        return f"line {line_no}, {repr(pref)} ^ {repr(suff)}"


class ParseError(Exception):
    def __init__(self, pos: Position, message: str):
        self.pos = pos
        super().__init__(f"{message} at {pos}")


# AST Node Classes
class ASTNode(ABC):
    @abstractmethod
    def __str__(self) -> str:
        pass

    def is_atomic(self) -> bool:
        return True

    def is_terminal(self) -> bool:
        return all(c.is_terminal() for c in self.children())

    def top_str(self) -> str:
        return self.__str__()

    def simplify(self) -> "ASTNode":
        return self

    def children(self) -> list["ASTNode"]:
        return []


@dataclass
class LiteralNode(ASTNode):
    value: str

    def __str__(self) -> str:
        return f'"{self.value}"'


@dataclass
class RegexNode(ASTNode):
    rx: str

    def __str__(self) -> str:
        return f"/{self.rx}/"


@dataclass
class RuleRefNode(ASTNode):
    name: str
    target: Optional["RuleNode"] = None

    def is_terminal(self) -> bool:
        if self.target is None:
            return False
        return self.target.rule_is_terminal

    def __str__(self) -> str:
        if self.target is None:
            return self.name
        return self.target.name


@dataclass
class RepetitionNode(ASTNode):
    node: ASTNode
    min_times: int
    max_times: Optional[int]  # None represents unlimited

    def children(self) -> List[ASTNode]:
        return [self.node]

    def simplify(self) -> ASTNode:
        self.node = self.node.simplify()
        return self

    def __str__(self) -> str:
        inner = str(self.node)
        if not self.node.is_atomic():
            inner = f"({inner})"
        if self.min_times == 0 and self.max_times is None:
            return f"{inner}*"
        if self.min_times == 1 and self.max_times is None:
            return f"{inner}+"
        if self.min_times == 0 and self.max_times == 1:
            return f"{inner}?"
        max_str = str(self.max_times) if self.max_times is not None else ""
        return f"{inner}{{{self.min_times},{max_str}}}"


@dataclass
class SequenceNode(ASTNode):
    nodes: List[ASTNode]

    def __str__(self) -> str:
        if not self.nodes:
            return '""'
        return " ".join(str(node) for node in self.nodes)

    def is_atomic(self) -> bool:
        return False

    def simplify(self) -> ASTNode:
        for i in range(len(self.nodes)):
            self.nodes[i] = self.nodes[i].simplify()
        if len(self.nodes) == 1:
            return self.nodes[0]
        return self

    def children(self) -> list[ASTNode]:
        return self.nodes


@dataclass
class AlternativeNode(ASTNode):
    alternatives: List[ASTNode]

    def top_str(self) -> str:
        return "\n     | ".join(str(alt) for alt in self.alternatives)

    def __str__(self) -> str:
        return "(" + " | ".join(str(alt) for alt in self.alternatives) + ")"

    def is_atomic(self) -> bool:
        return False

    def simplify(self) -> ASTNode:
        for i in range(len(self.alternatives)):
            self.alternatives[i] = self.alternatives[i].simplify()
        if len(self.alternatives) == 1:
            return self.alternatives[0]
        return self

    def children(self) -> list[ASTNode]:
        return self.alternatives


@dataclass
class RuleNode(ASTNode):
    name: str
    alternatives: ASTNode
    comment: str
    rule_is_terminal: bool = False
    order = 0

    def children(self) -> List[ASTNode]:
        return [self.alternatives]

    def __str__(self) -> str:
        return f"{self.comment}{self.name}: {self.alternatives.top_str()}"


class GrammarParser:
    def __init__(self) -> None:
        self.curr_comment = ""
        pass

    def parse(self, text: str) -> dict[str, RuleNode]:
        pos = Position(text, 0)
        pos = self._skip_space(pos, allow_newlines=True)
        rules: list[RuleNode] = []

        while pos.current():
            rule, pos = self._parse_rule(pos)
            rules.append(rule)
            pos = self._skip_space(pos, allow_newlines=True)

        return {rule.name: rule for rule in rules}

    def _parse_char(self, pos: Position) -> Tuple[str, Position]:
        def is_all_hex(s: str) -> bool:
            return all(ch in "0123456789abcdefABCDEF" for ch in s)

        if pos.current() == "\\":
            if not pos.peek(2)[1]:
                raise ParseError(pos, "Incomplete escape sequence")
            pos = pos.advance()
            c = pos.current()
            if c in '"\\[]nrt':
                return "\\" + c, pos.advance()
            elif c == "x":
                hex_value = pos.peek(3)[1:3]
                if len(hex_value) != 2 or not is_all_hex(hex_value):
                    raise ParseError(
                        pos, f"Invalid \\x escape sequence: \\x{hex_value}"
                    )
                pos = pos.advance(3)
                return f"\\x{hex_value}", pos
            elif c == "u":
                hex_value = pos.peek(5)[1:5]
                if len(hex_value) != 4 or not is_all_hex(hex_value):
                    raise ParseError(
                        pos, f"Invalid \\u escape sequence: \\u{hex_value}"
                    )
                pos = pos.advance(5)
                return f"\\u{hex_value.lstrip('0')}", pos
            elif c == "U":
                hex_value = pos.peek(9)[1:9]
                if len(hex_value) != 8 or not is_all_hex(hex_value):
                    raise ParseError(
                        pos, f"Invalid \\U escape sequence: \\U{hex_value}"
                    )
                pos = pos.advance(9)
                return f"\\U{hex_value.lstrip('0')}", pos
            else:
                raise ParseError(pos, f"Invalid escape sequence \\{c}")
        elif pos.current() == "":
            raise ParseError(pos, "Unexpected end of input")

        return pos.current(), pos.advance()

    def _parse_char_class(self, pos: Position) -> Tuple[ASTNode, Position]:
        if pos.current() != "[":
            raise ParseError(pos, "Expected '['")
        r = "["
        pos = pos.advance()

        while True:
            c, pos = self._parse_char(pos)
            if c in "/[":
                r += "\\" + c
            else:
                r += c
            if c == "]":
                break

        return RegexNode(r), pos

    def _parse_literal(self, pos: Position) -> Tuple[ASTNode, Position]:
        if pos.current() != '"':
            raise ParseError(pos, "Expected '\"'")
        pos = pos.advance()
        r = ""

        while True:
            c, pos = self._parse_char(pos)
            if c == '"':
                break
            r += c

        return LiteralNode(r), pos

    @staticmethod
    def _parse_name(pos: Position) -> Tuple[str, Position]:
        start = pos.pos
        while GrammarParser._is_word_char(pos.current()):
            pos = pos.advance()
        if pos.pos == start:
            raise ParseError(pos, "Expected name")
        return pos.text[start : pos.pos], pos

    @staticmethod
    def _parse_int(pos: Position) -> Tuple[int, Position]:
        start = pos.pos
        while pos.current().isdigit():
            pos = pos.advance()
        if pos.pos == start:
            raise ParseError(pos, "Expected integer")
        return int(pos.text[start : pos.pos]), pos

    def _skip_space(self, pos: Position, allow_newlines: bool) -> Position:
        while pos.current():
            if pos.current() in " \t":
                pos = pos.advance()
            elif allow_newlines and pos.current() in "\r\n":
                pos = GrammarParser._skip_newline(pos)
            elif pos.current() == "#":
                pos = pos.advance()
                cmt = "//"
                while pos.current() and pos.current() not in "\r\n":
                    cmt += pos.current()
                    pos = pos.advance()
                self.curr_comment += cmt + "\n"
            else:
                break
        return pos

    @staticmethod
    def _skip_newline(pos: Position) -> Position:
        if pos.current() == "\r":
            pos = pos.advance()
            if pos.current() == "\n":
                pos = pos.advance()
        elif pos.current() == "\n":
            pos = pos.advance()
        return pos

    @staticmethod
    def _is_word_char(c: str) -> bool:
        return c.isalnum() or c == "-"

    def _parse_rule(self, pos: Position) -> Tuple[RuleNode, Position]:
        name, pos = self._parse_name(pos)
        pos = self._skip_space(pos, allow_newlines=False)

        if pos.peek(3) != "::=":
            raise ParseError(pos, "Expected ::=")
        pos = pos.advance(3)

        pos = self._skip_space(pos, allow_newlines=True)
        alternatives, pos = self._parse_alternatives(pos, is_nested=False)

        pos = self._skip_newline(pos)
        cmt = self.curr_comment
        self.curr_comment = ""
        return RuleNode(name, alternatives, cmt), pos

    def _parse_alternatives(
        self, pos: Position, is_nested: bool
    ) -> Tuple[AlternativeNode, Position]:
        alternatives: list[ASTNode] = []

        while True:
            sequence, pos = self._parse_sequence(pos, is_nested)
            alternatives.append(sequence)

            pos = self._skip_space(pos, allow_newlines=is_nested)
            if pos.current() != "|":
                break

            pos = pos.advance()
            pos = self._skip_space(pos, allow_newlines=True)

        return AlternativeNode(alternatives), pos

    def _parse_sequence(
        self, pos: Position, is_nested: bool
    ) -> Tuple[SequenceNode, Position]:
        nodes: List[ASTNode] = []

        while (
            pos.current()
            and pos.current() not in "|)"
            and (is_nested or pos.current() not in "\r\n")
        ):
            if pos.current() == '"':
                node, pos = self._parse_literal(pos)
                nodes.append(node)
            elif pos.current() == "[":
                node, pos = self._parse_char_class(pos)
                nodes.append(node)
            elif pos.current() == "(":
                node, pos = self._parse_group(pos, is_nested=is_nested)
                nodes.append(node)
            elif pos.current() == ".":
                nodes.append(RegexNode("."))
                pos = pos.advance()
            elif self._is_word_char(pos.current()):
                name, pos = self._parse_name(pos)
                nodes.append(RuleRefNode(name))
            else:
                break

            pos = self._skip_space(pos, allow_newlines=is_nested)
            pos = self._parse_repetition(pos, nodes)
            pos = self._skip_space(pos, allow_newlines=is_nested)

        return SequenceNode(nodes), pos

    def _parse_group(self, pos: Position, is_nested: bool) -> Tuple[ASTNode, Position]:
        if pos.current() != "(":
            raise ParseError(pos, "Expected '('")
        pos = pos.advance()
        pos = self._skip_space(pos, True)

        alternatives, pos = self._parse_alternatives(pos, is_nested=True)

        if pos.current() != ")":
            raise ParseError(pos, "Expected ')'")
        pos = pos.advance()

        return alternatives, self._skip_space(pos, is_nested)

    def _parse_repetition(self, pos: Position, nodes: List[ASTNode]) -> Position:
        if not nodes:
            return pos

        if pos.current() == "*":
            nodes[-1] = RepetitionNode(nodes[-1], 0, None)
            return pos.advance()
        elif pos.current() == "+":
            nodes[-1] = RepetitionNode(nodes[-1], 1, None)
            return pos.advance()
        elif pos.current() == "?":
            nodes[-1] = RepetitionNode(nodes[-1], 0, 1)
            return pos.advance()
        elif pos.current() == "{":
            pos = pos.advance()
            pos = self._skip_space(pos, True)
            min_times, pos = self._parse_int(pos)
            pos = self._skip_space(pos, True)

            if pos.current() == "}":
                nodes[-1] = RepetitionNode(nodes[-1], min_times, min_times)
                return pos.advance()
            elif pos.current() == ",":
                pos = self._skip_space(pos.advance(), True)
                max_times = None
                if pos.current().isdigit():
                    max_times, pos = self._parse_int(pos)
                pos = self._skip_space(pos, True)
                if pos.current() != "}":
                    raise ParseError(pos, "Expected '}'")
                nodes[-1] = RepetitionNode(nodes[-1], min_times, max_times)
                return pos.advance()
            else:
                raise ParseError(pos, "Expected ',' or '}'")

        return pos


def resolve(rules: dict[str, RuleNode]) -> None:
    def rename(r: RuleNode, name: str) -> None:
        if name in rules:
            raise Exception(f"Rule '{name}' already exists")
        del rules[r.name]
        r.name = name
        rules[name] = r

    for i, r in enumerate(rules.values()):
        r.order = i
        r.alternatives = r.alternatives.simplify()

    def all_children(node: ASTNode) -> Iterator[ASTNode]:
        for c in node.children():
            yield c
            yield from all_children(c)

    for r in rules.values():
        for node in all_children(r):
            if isinstance(node, RuleRefNode):
                if node.name not in rules:
                    raise Exception(f"Rule '{node.name}' not found")
                node.target = rules[node.name]

    if "root" not in rules:
        raise Exception("No 'root' rule found")
    rename(rules["root"], "start")

    num_fix = 1
    while num_fix > 0:
        num_fix = 0
        for r in rules.values():
            if (
                r.name != "start"
                and not r.rule_is_terminal
                and r.alternatives.is_terminal()
            ):
                r.rule_is_terminal = True
                num_fix += 1

    for r in list(rules.values()):
        new_name = r.name.replace("-", "_")
        # convert fooBar_Baz to foo_bar_baz
        new_name = re.sub(r"([a-z])([A-Z])", r"\1_\2", new_name).lower()
        if r.rule_is_terminal:
            new_name = new_name.upper()
        else:
            new_name = new_name.lower()
        if r.name != new_name:
            rename(r, new_name)


def gbnf_to_lark(text: str) -> str:
    """
    Convert a GBNF (llama.cpp) grammar to Lark syntax.
    """
    parser = GrammarParser()
    rules = parser.parse(text)
    resolve(rules)
    rlist = list(rules.values())
    rlist.sort(key=lambda r: r.order)
    res = "%llguidance {}\n\n"
    prev_nl = True
    for r in rlist:
        s = str(r)
        if not prev_nl and "\n" in s:
            res += "\n"
        res += s + "\n"
        prev_nl = "\n" in s
        if prev_nl:
            res += "\n"
    return res


def is_lark_syntax(text: str) -> bool:
    """
    Check if the text is already in Lark syntax.
    """
    return re.search(r"(?m)^\s*(%llguidance\b|start\s*:)", text) is not None


def any_to_lark(text: str) -> str:
    """
    Convert a grammar to Lark syntax if needed.
    """
    if is_lark_syntax(text):
        return text
    return gbnf_to_lark(text)


def main() -> None:
    import sys
    import os

    def process_file(fn: str) -> None:
        print(f"{fn}... ", end="", flush=True)
        with open(fn) as f:
            text = f.read()
        lark = gbnf_to_lark(text)
        fn_lark = os.path.splitext(fn)[0] + ".lark"
        with open(fn_lark, "w") as f:
            f.write(lark)
        print(f"OK")

    if len(sys.argv) < 2:
        print("Usage: gbnf_to_lark.py <file1> <file2> ...")
        sys.exit(1)
    for fn in sys.argv[1:]:
        process_file(fn)


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import json
import sys

"""
To use this, run Guidance with LLGUIDANCE_TEST_TRACE=1 LLGUIDANCE_LOG_LEVEL=2 and
then pass the output here.
"""

def testcase_from_logs(logs: str):
    sep = "‧"
    pairs = []
    prev_res = None
    prompt = None
    for line in logs.split("\n"):
        if line.startswith("TEST: "):
            obj = json.loads(line[6:])
            if prompt is None:
                prompt = obj["res_prompt"]
                continue
            if "res_prompt" in obj:
                print("There's a second testcase here")
                break
            if prev_res:
                pairs.append((prev_res, obj["arg"]))
            prev_res = obj["res"]
    # assert prev_res == "stop"
    testcase = [prompt]
    gen_tokens = []

    def flush_gen_tokens():
        testcase.append(sep.join(gen_tokens))
        gen_tokens.clear()

    for res, arg in pairs:
        print(res, arg)
        if res["sample_mask"]:
            gen_tokens.append(arg["tokens"])
        else:
            splice = res["splices"][0]
            t0 = splice["tokens"]
            assert t0 == arg["tokens"]
            flush_gen_tokens()
            if splice["backtrack"]:
                t0 = str(splice["backtrack"]) + "↶" + t0
            testcase.append(t0)
    if gen_tokens:
        flush_gen_tokens()

    print("Testcase:", testcase)


def main():
    with open(sys.argv[1], "r") as f:
        logs = f.read()
    testcase_from_logs(logs)

main()



#!/usr/bin/env python3

import os
import glob
import json

splits = [
    "Github_trivial",
    "Github_easy",
    "Github_hard",
    "Github_medium",
    "Github_ultra",
    "Glaiveai2K",
    "Kubernetes",
    "Snowplow",
    "WashingtonPost",
    "MCPspec",
    "JsonSchemaStore",
    # "Handwritten",
    # "Synthesized",
    "TOTAL",
]


def markdown_table(data):
    data = [[str(e) for e in l] for l in data]
    col_widths = [max(len(row[col]) for row in data) for col in range(len(data[0]))]

    # Generate the Markdown table
    def format_row(row):
        return (
            "| "
            + " | ".join(
                f"{cell:<{col_widths[i]}}" if i == 0 else f"{cell:>{col_widths[i]}}"
                for i, cell in enumerate(row)
            )
            + " |\n"
        )

    r = format_row(data[0])
    r += (
        "|"
        + "|".join(
            ":" + "-" * (width + 1) if i == 0 else "-" * (width + 1) + ":"
            for i, width in enumerate(col_widths)
        )
        + "|\n"
    )

    # Print the rows
    for row in data[1:]:
        r += format_row(row)

    return r


def perc(a, b):
    return f"{a/b:.0%}" if b > 0 else "N/A"


def json_markdown_table(data: list[dict]):
    hd = list(data[0].keys())
    rows = [hd]
    for d in data:
        row = [d[k] for k in hd]
        rows.append(row)
    return markdown_table(rows)


def main():
    stats_by_id: dict[str, dict] = {}
    for f in glob.glob("tmp/splits/*.json"):
        id = os.path.basename(f).split(".")[0]
        with open(f, "r") as file:
            data = json.load(file)
            stats_by_id[id] = data
    # splits = [s.split("-")[0] for s in stats_by_id.keys()]
    # splits = list(set(splits))
    # splits.sort()
    rows = []
    for s in splits:
        valid = stats_by_id[f"{s}-valid"]
        invalid = stats_by_id[f"{s}-invalid"]
        compact = stats_by_id[f"{s}-valid-compact"]
        row = {
            "split": s,
            "schemas": valid["num_files"],
            "has tests": perc(
                valid["num_files"] - valid["num_testless_files"], valid["num_files"]
            ),
            "valid inst.": valid["num_valid_tests"],
            "invalid inst.": invalid["num_invalid_tests"],
            "tok/inst.": (
                valid["llg_json"]["num_all_tokens"] // valid["num_valid_tests"]
            ),
            "FF": perc(valid["llg"]["ff_fraction"], 1.0),
            "FF compact": perc(compact["llg"]["ff_fraction"], 1.0),
        }
        llg_row = {
            "TTFM us": valid["llg"]["ttfm_us"],
            "TBT us": valid["llg"]["mask_us"],
            "pass": perc(
                valid["llg"]["num_correct_schemas"],
                valid["num_files"],
            ),
        }
        rows.append(
            {
                **row,
                # **llg_row,
            }
        )
    print(json_markdown_table(rows))


if __name__ == "__main__":
    main()



from ._lib import LLMatcher
from typing import Literal
from .gbnf_to_lark import any_to_lark

GrammarFormat = Literal[
    "lark",
    "gbnf",
    "ebnf",
    "cfg",
    "grammar",
    "json_schema",
    "json",
    "regex",
    "llguidance",
]


def grammar_from(format: GrammarFormat, text: str) -> str:
    """
    Create a llguidance grammar definition from a given grammar text
    of the specified type.

    Args:
        format: The format of the grammar text
            "lark": Lark grammar, see https://github.com/guidance-ai/llguidance/blob/main/docs/syntax.md
            "gbnf", "ebnf", "cfg", "grammar": Lark grammar or GBNF grammar, see https://github.com/ggml-org/llama.cpp/blob/master/grammars/README.md
            "json_schema", "json": JSON schema, see https://github.com/guidance-ai/llguidance/blob/main/docs/json_schema.md
            "regex": Regular expression, see https://docs.rs/regex/latest/regex/#syntax
            "llguidance": JSON object like: {"grammars": [{"lark_grammar": "..."},{"json_schema": {...}}]}
        text: The grammar text

    Returns:
        The llguidance grammar definition as a string.
        This can be passed to LLInterpreter or LLMatcher.

    Raises:
        ValueError: If the format is not recognized

    Note:
        To get "any JSON object" grammar use:
            grammar_from("json_schema", '{"type": "object"}')
    """

    if format == "lark":
        return LLMatcher.grammar_from_lark(text)
    if format in ("gbnf", "ebnf", "cfg", "grammar"):
        return LLMatcher.grammar_from_lark(any_to_lark(text))
    if format in ("json_schema", "json"):
        return LLMatcher.grammar_from_json_schema(text)
    if format == "regex":
        return LLMatcher.grammar_from_regex(text)
    if format == "llguidance":
        return text
    raise ValueError(f"Unknown grammar type: {format}")



#!/usr/bin/env python3

import re
import subprocess
import sys
import os

pyproject_path = "pyproject.toml"
cargo_paths = ["parser", "python_ext", "toktrie", "toktrie_hf_tokenizers"]
version_pattern = r'\nversion\s*=\s*"(\d+\.\d+\.\d+)([^"]*)"'


def get_current_version(file_path):
    with open(file_path, "r") as f:
        content = f.read()
    match = re.search(version_pattern, content)
    if match:
        return match.group(1)
    raise ValueError(f"Version not found in {file_path}")


def bump_patch_version(version: str):
    major, minor, patch = map(int, version.split("."))
    patch += 1
    return f"{major}.{minor}.{patch}"


def update_version_in_file(file_path, new_version):
    with open(file_path, "r") as f:
        content = f.read()

    new_content = re.sub(version_pattern, f'\nversion = "{new_version}"',
                         content)

    with open(file_path, "w") as f:
        f.write(new_content)


def check_in_and_tag(version):
    subprocess.run(["git", "add", pyproject_path] +
                   [p + "/Cargo.toml" for p in cargo_paths] +
                   ["Cargo.lock"],
                   check=True)
    subprocess.run(["git", "commit", "-m", f"Bump version to {version}"],
                   check=True)
    subprocess.run(["git", "tag", f"v{version}"], check=True)
    subprocess.run(["git", "push"], check=True)
    subprocess.run(["git", "push", "--tags"], check=True)


def ensure_clean_working_tree():
    status_output = subprocess.run(["git", "status", "--porcelain"],
                                   capture_output=True,
                                   text=True).stdout
    if status_output.strip():
        subprocess.run(["git", "status"])
        print(
            "\n\nWorking tree is not clean. Please commit or stash your changes before running this script.\n"
        )
        sys.exit(1)


def main():
    #subprocess.run(["python3", "./scripts/update-git.py"], check=True)

    ensure_clean_working_tree()

    current_version = get_current_version(pyproject_path)
    suggested_version = bump_patch_version(current_version)

    print(f"Current version: {current_version}")
    new_version = (input(f"Enter new version (default: {suggested_version}): ")
                   or suggested_version)

    update_version_in_file(pyproject_path, new_version.replace("-", ""))

    for p in cargo_paths:
        update_version_in_file(p + "/Cargo.toml", new_version)
    subprocess.run(["cargo", "check"], check=True)

    check_in_and_tag(new_version)


if __name__ == "__main__":
    main()



from ._lib import (
    LLTokenizer,
    LLInterpreter,
    JsonCompiler,
    LarkCompiler,
    RegexCompiler,
    LLExecutor,
    LLMatcher,
)
from ._tokenizer import TokenizerWrapper
from ._grammar_from import GrammarFormat, grammar_from

__all__ = [
    "LLTokenizer",
    "LLMatcher",
    "LLInterpreter",
    "LLExecutor",
    "JsonCompiler",
    "LarkCompiler",
    "RegexCompiler",
    "TokenizerWrapper",
    "grammar_from",
    "GrammarFormat",
]



#!/usr/bin/env python3

# For standalone use, please use this:
# https://github.com/guidance-ai/llguidance/blob/main/python/llguidance/gbnf_to_lark.py

import os
import subprocess
import sys

p = os.path.join(os.path.dirname(__file__), "../python/llguidance/gbnf_to_lark.py")
r = subprocess.run(["python3", p] + sys.argv[1:], stdout=sys.stdout, stderr=sys.stderr)
sys.exit(r.returncode)



from typing import List, Optional
from ._lib import LLTokenizer

import transformers # type: ignore[import-untyped]


def from_tokenizer(
    hf_tokenizer: transformers.PreTrainedTokenizerBase,
    n_vocab: Optional[int] = None,
    eos_token: Optional[int] = None,
    slices: Optional[List[str]] = None,
) -> LLTokenizer:
    """
    Create a new tokenizer from a Hugging Face tokenizer.
    This is an expensive operation (~1s), so the result should be cached.
    It also currently creates a non-canonical tokenizer, which means it cannot
    produce fast-forward tokens (though it can produce fast-forward bytes).

    Args:
        hf_tokenizer: transformers.PreTrainedTokenizerBase - the tokenizer to wrap
        n_vocab: int - override the size of the vocabulary
        eos_token: int - override the EOS token
        slices: List[str] - configuration for slicer optimization; pass [] to disable,
            or None to use the default configuration
    """

    if isinstance(hf_tokenizer, transformers.PreTrainedTokenizerFast):
        # this is not ideal...
        s = hf_tokenizer.backend_tokenizer.to_str()
        if n_vocab is None:
            n_vocab = hf_tokenizer.vocab_size
        if eos_token is None:
            eos_token = hf_tokenizer.eos_token_id
        return LLTokenizer(s,
                           n_vocab=n_vocab,
                           eos_token=eos_token,
                           slices=slices)
    else:
        raise ValueError("Only fast tokenizers are supported")



from typing import List, Tuple, Dict, Any, Optional, Callable

import torch
import numpy as np
import pytest
import json
import time

from llguidance.torch import (
    apply_token_bitmask_inplace,
    get_bitmask_shape,
    fill_next_token_bitmask,
    allocate_token_bitmask,
    fill_next_token_bitmask_par,
)
from llguidance import LLMatcher, LLTokenizer, LLExecutor

import llguidance.hf
import transformers  # type: ignore[import-untyped]


def _build_tokenizer() -> LLTokenizer:
    hf_tok = transformers.AutoTokenizer.from_pretrained(
        "unsloth/Meta-Llama-3.1-8B-Instruct")
    return llguidance.hf.from_tokenizer(hf_tok)


_tokenizer: Optional[LLTokenizer] = None


def tokenizer() -> LLTokenizer:
    global _tokenizer
    if _tokenizer is None:
        _tokenizer = _build_tokenizer()
    return _tokenizer


def lark_matcher(grm: str) -> LLMatcher:
    gstr = json.dumps({"grammars": [{"lark_grammar": grm}]})
    interp = LLMatcher(tokenizer(), gstr, log_level=1)
    return interp


def test_grammar() -> None:
    t = tokenizer()
    mask = allocate_token_bitmask(2, t.vocab_size)
    interp = lark_matcher(r"start: /[A-Z ]*/")
    fill_next_token_bitmask(interp, mask)
    allowed = []
    for idx, v in enumerate(mask[0, :].tolist()):
        for bit_idx in range(32):
            tok_idx = idx * 32 + bit_idx
            if v & (1 << bit_idx):
                if t.is_special_token(tok_idx):
                    continue
                s = t.decode_str([tok_idx])
                for c in s:
                    assert c.isupper() or c.isspace()
                allowed.append(tok_idx)
    assert len(allowed) > 100
    interp.consume_token(allowed[3])
    fill_next_token_bitmask(interp, mask, 1)
    assert torch.isclose(mask[1, :], mask[0, :]).all()


def test_par_grammar() -> None:
    n_gram = 50
    t = tokenizer()
    grammars = [(lark_matcher(r"start: /[a-zA-Z ]*/"), idx)
                for idx in range(n_gram)]
    mask = allocate_token_bitmask(n_gram, t.vocab_size)
    mask2 = allocate_token_bitmask(n_gram, t.vocab_size)
    exec = LLExecutor()
    t0 = time.monotonic()
    fill_next_token_bitmask_par(exec, grammars, mask)
    par_time = int((time.monotonic() - t0) * 1_000_000)
    for i in range(n_gram):
        assert torch.isclose(mask[i, :], mask[0, :]).all()
    t0 = time.monotonic()
    for g, idx in grammars:
        fill_next_token_bitmask(g, mask2, idx)
    seq_time = int((time.monotonic() - t0) * 1_000_000)
    assert torch.isclose(mask, mask2).all()
    print(f"Parallel: {par_time} us, Sequential: {seq_time} us")


def asserts(msg: str, fn: Callable[..., Any], *args: Any) -> None:
    try:
        fn(*args)
        raise Exception("OK")
    except Exception as e:
        if msg in str(e):
            return
        raise AssertionError(f"Expected {msg}, got {e}")


if __name__ == "__main__":
    test_par_grammar()



#!/usr/bin/env python3

import os
import re
import subprocess

def get_toktrie_version():
    """Extracts the version of toktrie from its Cargo.toml using regex."""
    cargo_toml_path = os.path.join("toktrie", "Cargo.toml")
    with open(cargo_toml_path, "r") as f:
        content = f.read()

    match = re.search(r'^version\s*=\s*"(.*?)"', content, re.MULTILINE)
    if not match:
        raise ValueError("Could not find toktrie version in Cargo.toml")
    
    return match.group(1)

def update_dependency(crate, version):
    """Replaces `workspace = true` dependency in Cargo.toml with actual version."""
    cargo_toml_path = os.path.join(crate, "Cargo.toml")
    with open(cargo_toml_path, "r") as f:
        content = f.read()

    updated_content = re.sub(
        r'toktrie\s*=\s*\{ workspace = true \}',
        f'toktrie = {{ version = "{version}" }}',
        content
    )

    with open(cargo_toml_path, "w") as f:
        f.write(updated_content)

    return content  # Return original content for restoration

def restore_dependency(crate, original_content):
    """Restores the original Cargo.toml content."""
    cargo_toml_path = os.path.join(crate, "Cargo.toml")
    with open(cargo_toml_path, "w") as f:
        f.write(original_content)

def publish_crate(crate):
    """Runs `cargo publish` in the specified crate directory."""
    subprocess.run(["cargo", "publish", "--allow-dirty"], cwd=crate, check=True)

def main():
    toktrie_version = get_toktrie_version()
    
    # Publish toktrie first
    print(f"Publishing toktrie v{toktrie_version}...")
    publish_crate("toktrie")

    # Publish dependent crates
    for crate in ["toktrie_hf_tokenizers", "parser"]:
        print(f"Updating {crate} to use toktrie v{toktrie_version}...")
        original_content = update_dependency(crate, toktrie_version)

        try:
            print(f"Publishing {crate}...")
            publish_crate(crate)
        finally:
            print(f"Restoring original {crate} Cargo.toml...")
            restore_dependency(crate, original_content)

    print("All crates published successfully.")

if __name__ == "__main__":
    main()


from typing import Any, Dict, List, Tuple
import llguidance
from llguidance.numpy import fill_next_token_bitmask_par, allocate_token_bitmask
import pytest
from numpy.typing import NDArray
import numpy as np

_tokenizer = None


def tokenizer() -> llguidance.LLTokenizer:
    global _tokenizer
    if _tokenizer is None:
        _tokenizer = llguidance.LLTokenizer("byte")
    return _tokenizer


def matcher(grm: str) -> llguidance.LLMatcher:
    return llguidance.LLMatcher(tokenizer(), grm, log_level=1)


def check_one_grammar(grm: str, s: str, passing: bool) -> None:
    # print("Checking", repr(s))
    interp = matcher(grm)
    final_reject = False
    if s.startswith("FINAL_REJECT:"):
        final_reject = True
        s = s[len("FINAL_REJECT:"):]
    tokens = tokenizer().tokenize_str(s)
    for i, t in enumerate(tokens):
        next_tokens = tokens[i:]
        if passing or final_reject:
            assert interp.validate_tokens(next_tokens) == len(next_tokens)
        else:
            assert interp.validate_tokens(next_tokens) < len(next_tokens)
        mask = interp.compute_logit_bias()
        if mask[t] == 0:
            if passing or final_reject:
                print("Token not in mask",
                      tokenizer().dbg_tokens(tokens[:i + 1]), repr(s))
                assert False
            return
        else:
            assert mask[t] == 200
        interp.consume_token(t)
    if final_reject:
        if interp.is_accepting():
            print("Expected to fail at the end", s)
            assert False
        else:
            return
    if not passing:
        print("Expected to fail", s)
        assert False
    assert interp.is_accepting()


def check_grammar(grm: str, passing: List[str], failing: List[str]) -> None:
    for s in passing:
        check_one_grammar(grm, s, True)
    for s in failing:
        check_one_grammar(grm, s, False)


def test_json() -> None:
    grm = llguidance.LLMatcher.grammar_from_json_schema(
        {"type": "object"}, {"whitespace_flexible": False})
    check_grammar(grm, ["{}", '{"foo":1}'], ["FINAL_REJECT:{", " {}", "{ }"])

    grm = llguidance.LLMatcher.grammar_from_json_schema({
        "type": "object",
        "properties": {
            "foo": {
                "type": "integer"
            }
        },
        "required": ["foo"]
    })
    check_grammar(grm, ['{"foo":1}', '{"foo":1,"bar":2}', '{ "foo" : 1 }'],
                  ["{}", "FINAL_REJECT:{", ' {"foo":1}', '{"bar":1}'])


def test_lark() -> None:
    check_grammar(
        'start: /.../ "abc" /.../',
        [
            "abcabcabc",
            "aaaabcccc",
            # NOTE: Also ensures that multi-byte characters still count as a single character
            "🔵🟠✅abc❌🟠🔵",
        ],
        [
            "aaabcccc",
            "aaaaabcccc",
            "FINAL_REJECT:aaaabccc",
            "aaaabccccc",
            "🔵🟠✅❌abc❌✅🟠🔵",
            "🔵🟠abc🟠🔵",
        ],
    )


def test_regex_grammar() -> None:
    grm = llguidance.LLMatcher.grammar_from_regex(r"\d+")
    check_grammar(grm, ["123", "456"], ["abc", "1a2"])


def test_lark_syntax() -> None:
    with pytest.raises(ValueError, match="no_such_rule"):
        matcher('start: /.../ no_such_rule')


def test_slices() -> None:
    t = tokenizer()
    gen_slices = llguidance.LLTokenizer.general_slices()
    assert len(gen_slices) > 0
    json_slices = llguidance.LLTokenizer.json_slices()
    assert len(json_slices) > 0
    t2 = t.with_slices(json_slices)
    assert t.tokenize_str("Hello, world!") == t2.tokenize_str("Hello, world!")


def mask_has(mask: NDArray[np.int32], t: int) -> bool:
    v: int = mask[t // 32]
    return v & (1 << (t % 32)) != 0


def test_par_errors() -> None:
    t = tokenizer()
    exec = llguidance.LLExecutor()
    g0 = matcher(r"start: /[a-zA-Z ]*/")
    g1 = matcher(r"start: /[0-9 ]*/")
    mask = allocate_token_bitmask(3, t.vocab_size)

    with pytest.raises(ValueError, match="Target index out of bounds"):
        fill_next_token_bitmask_par(exec, [(g0, 0), (g1, 3)], mask)

    with pytest.raises(RuntimeError, match="Already borrowed"):
        fill_next_token_bitmask_par(exec, [(g0, 0), (g1, 1), (g1, 2)], mask)

    with pytest.raises(TypeError, match="cannot be converted"):
        l = [1, (g1, 0), (g1, 1)]
        fill_next_token_bitmask_par(exec, l, mask)  # type: ignore

    with pytest.raises(TypeError, match="cannot be converted"):
        l = [(tokenizer(), 0)]
        fill_next_token_bitmask_par(exec, l, mask)  # type: ignore

    with pytest.raises(ValueError, match=r"Expecting.*tuple"):
        l = [(tokenizer(), 0, 0)]
        fill_next_token_bitmask_par(exec, l, mask)  # type: ignore

    (three, vocab) = mask.shape
    assert three == 3
    with pytest.raises(ValueError, match="Null pointer"):
        exec.unsafe_compute_mask_ptr([(g0, 0), (g1, 1)], 0, vocab * 4, 3)
    with pytest.raises(ValueError, match="Pointer not aligned"):
        exec.unsafe_compute_mask_ptr([(g0, 0), (g1, 1)], 3, vocab * 4, 3)
    with pytest.raises(ValueError, match="Invalid buffer size"):
        exec.unsafe_compute_mask_ptr([(g0, 0), (g1, 1)], 1024, vocab * 4 + 1,
                                     3)
    with pytest.raises(ValueError, match="Invalid buffer size"):
        exec.unsafe_compute_mask_ptr([(g0, 0), (g1, 1)], 1024, vocab * 4 - 1,
                                     3)

    # should be OK
    fill_next_token_bitmask_par(exec, [(g0, 0), (g1, 2)], mask)
    t_a = t.tokenize_str("a")[0]
    t_1 = t.tokenize_str("1")[0]
    assert mask_has(mask[0, :], t_a)
    assert not mask_has(mask[0, :], t_1)
    assert not mask_has(mask[2, :], t_a)
    assert mask_has(mask[2, :], t_1)


def consume_tokens(m: llguidance.LLMatcher, tokens: List[int]) -> None:
    print("Consume", tokenizer().dbg_tokens(tokens))
    assert m.stop_reason() == "NotStopped"
    assert not m.is_stopped()
    assert not m.is_accepting()
    for t in tokens:
        mask = m.compute_logit_bias()
        assert mask[t] == 200, "Token should be in mask."
        bit_mask = m.compute_bitmask()
        assert bit_mask[t // 8] & (1 <<
                                   (t % 8)) != 0, "Token should be in bitmask."
        assert m.stop_reason() == "NotStopped"
        assert not m.is_stopped()
        assert not m.is_accepting()
        m.consume_token(t)
    assert not m.is_error()


def test_stopping_conditions() -> None:
    m = llguidance.LLMatcher(tokenizer(), "start: /[aA][bB][cC]/")
    consume_tokens(m, tokenizer().tokenize_str("abc"))
    assert m.is_accepting()
    assert m.is_stopped()
    assert m.stop_reason() == "NoExtension"


def test_rollback() -> None:
    m = llguidance.LLMatcher(tokenizer(), "start: /[aA] [bB] [cC] [dD] [eE]/")
    m2 = m.deep_copy()
    t = tokenizer().tokenize_str("a b c d e")
    consume_tokens(m, t[0:3])
    assert not m.is_stopped() and not m.is_accepting()
    m.rollback(2)
    m3 = m.deep_copy()
    consume_tokens(m, t[1:])
    assert m.is_stopped() and m.is_accepting()
    assert not m.is_error()
    m.rollback(1)
    assert not m.is_accepting() and not m.is_stopped()
    assert m.stop_reason() == "NotStopped"
    consume_tokens(m, t[-1:])
    assert m.is_stopped() and m.is_accepting()
    assert not m.is_error()

    mask = m.compute_logit_bias()
    assert mask[tokenizer().eos_token] == 200
    assert not m.is_error()
    m.consume_token(tokenizer().eos_token)
    assert not m.is_error()

    consume_tokens(m2, t)
    assert m2.is_stopped() and m2.is_accepting() and not m2.is_error()

    m3.consume_tokens(t[1:])
    assert m2.is_stopped() and m2.is_accepting() and not m2.is_error()


def check_ff(m: llguidance.LLMatcher, expected: str) -> None:
    assert m.compute_ff_bytes() == expected.encode(), "FF bytes mismatch"
    assert m.compute_ff_tokens() == tokenizer().tokenize_str(expected)


def test_fast_forward() -> None:
    m = llguidance.LLMatcher(tokenizer(), "start: /(foo[12]23|bar)/")
    toks = tokenizer().tokenize_str("foo123")
    assert len(toks) == 6
    check_ff(m, "")
    consume_tokens(m, toks[0:1])
    check_ff(m, "oo")
    consume_tokens(m, toks[1:2])
    check_ff(m, "o")
    consume_tokens(m, toks[2:3])
    check_ff(m, "")
    consume_tokens(m, toks[3:4])
    check_ff(m, "23")
    consume_tokens(m, toks[4:])
    assert m.is_accepting()
    assert m.is_stopped()
    assert m.stop_reason() == "NoExtension"
    assert m.compute_ff_bytes() == b""
    assert m.compute_ff_tokens() == []
    assert not m.is_error()


def test_try_consume_tokens() -> None:
    m = llguidance.LLMatcher(tokenizer(), "start: /(foo[12]23|bar)/")
    tokens = tokenizer().tokenize_str("foo723")
    assert len(tokens) == 6
    assert m.try_consume_tokens(tokens) == 3
    consume_tokens(m, tokenizer().tokenize_str("123"))
    assert m.is_stopped() and m.is_accepting() and not m.is_error()


def test_consume_token_error() -> None:
    m = llguidance.LLMatcher(tokenizer(), "start: /(foo[12]23|bar)/")
    m2 = m.deep_copy()
    m3 = m.deep_copy()
    m4 = m.deep_copy()
    tokens = tokenizer().tokenize_str("foo723")

    consume_tokens(m, tokens[0:3])
    mask = m.compute_logit_bias()
    assert mask[tokens[3]] == 0
    r = m.consume_token(tokens[3])
    assert r == False
    assert m.is_error()
    assert "doesn't satisfy the grammar" in m.get_error()
    # this is internal error, since the token was not in the mask
    assert m.stop_reason() == "InternalError"

    consume_tokens(m2, tokens[0:3])
    r = m2.consume_token(tokenizer().vocab_size + 100)
    assert r == False
    assert m2.is_error()
    assert "out of range" in m2.get_error()

    r = m3.consume_tokens(tokens[0:3] + [tokenizer().vocab_size + 100])
    assert r == False
    assert m3.is_error()
    assert "out of range" in m3.get_error()

    n = m4.validate_tokens(tokens[0:3] + [tokenizer().vocab_size + 100])
    assert n == 0  # questionable
    assert m3.is_error()
    assert "out of range" in m3.get_error()



#!/usr/bin/env python3

# This script reads the output.map file and prints the sizes of the sections and modules.
# It works on macOS only.
# 
# Usage:
# RUSTFLAGS="-C link-arg=-Wl,-map,output.map" cargo build --release
# python3 rust_size.py output.map
# fx output.json

import sys
import subprocess
import os
import re
import random
import json

def run_command(command, input_data):
    process = subprocess.Popen(
        command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    stdout, stderr = process.communicate(input=input_data)
    return stdout, stderr


def main():
    if len(sys.argv) != 2:
        print(f"Usage: {sys.argv[0]} <output.map>")
        sys.exit(1)

    output_map = sys.argv[1]
    if not os.path.exists(output_map):
        print(f"Error: {output_map} does not exist")
        sys.exit(1)

    with open(output_map, "rb") as f:
        file_content = f.read()

    stdout, stderr = run_command(["rustfilt"], file_content)
    if stderr:
        print(f"Error: {stderr.decode('utf-8')}", file=sys.stderr)
        sys.exit(1)

    lib_by_id = {}
    sizes_by_lib = {}
    sizes_by_module = {}
    num_wrong = 0
    min_addr = None
    max_addr = 0
    sections = []

    for line in stdout.decode("utf-8").split("\n"):
        match = re.match(r"\[\s*(\d+)\] .*/([^/]+/[^-\.]+)", line) or re.match(
            r"\[\s*(\d+)\] ([^/]+)$", line
        )
        if match:
            id = int(match.group(1))
            lib = match.group(2)
            lib_by_id[id] = lib
            continue
        # 0x1001BEE8C	0x00007F58	__TEXT	__unwind_info
        match = re.match(
            r"(0x[0-9a-f]+)\s+(0x[0-9a-f]+)\s+\S+\s+__(\S+)", line, re.IGNORECASE
        )
        if match:
            start = int(match.group(1), 16)
            end = start + int(match.group(2), 16)
            sections.append((start, end, match.group(3)))
            continue

        # 0x100019778	0x00000180	[ 10] _core::ptr::drop_in_place<llguidance::api::GrammarWithLexer>
        match = re.match(
            r"(0x[0-9a-f]+)\s+(0x[0-9a-f]+)\s+\[\s*(\d+)\] (.+)", line, re.IGNORECASE
        )
        sys_modules = set(
            ["core", "std", "alloc", "collections", "panic_unwind", "rustc_demangle"]
        )
        if match:
            addr = int(match.group(1), 16)
            if min_addr is None or addr < min_addr:
                min_addr = addr
            max_addr = max(max_addr, addr)
            size = int(match.group(2), 16)
            id = int(match.group(3))
            name = match.group(4)
            if id not in lib_by_id:
                print(f"Error: lib not found for id {id}")
                continue
            lib = lib_by_id[id]
            section = "unk"
            for start, end, sec in sections:
                if start <= addr < end:
                    section = sec
                    break
            if lib not in sizes_by_lib:
                sizes_by_lib[lib] = {
                    "lib": lib,
                    "total": 0,
                    "sections": {},
                }
            e = sizes_by_lib[lib]
            if section not in e["sections"]:
                e["sections"][section] = 0
            e["sections"][section] += size
            e["total"] += size

            module = None
            for word in re.split(r"[<>&{}\s,\[\]\(\)\.]+", name):
                parts = list(w for w in word.split("::") if w)
                if len(parts) > 1:
                    parts[0] = re.sub(r"^_*", "", parts[0])
                    if (
                        parts[0] not in sys_modules
                        or module is None
                        or module[0] in sys_modules
                    ):
                        module = parts
            if module:
                d = sizes_by_module
                for part in module:
                    if part not in d:
                        d[part] = {"_": 0}
                    d[part]["_"] += size
                    d = d[part]

            continue
        if line.startswith("<<dead>>"):
            continue
        print(line)
        num_wrong += 1
        if num_wrong > 100:
            print("Too many wrong lines. STOP.")
            sys.exit(1)

    total_size = max_addr - min_addr

    entries = sorted(sizes_by_lib.values(), key=lambda x: x["total"], reverse=True)
    for e in entries:
        e["sections"] = dict(
            sorted(e["sections"].items(), key=lambda x: x[1], reverse=True)
        )

    cutoff_perc = 0.01
    cutoff = total_size / 100 * cutoff_perc

    byfile = {}
    for e in entries:
        lib = e["lib"]
        tsize = e["total"]
        if tsize < cutoff:
            continue
        k = f"{lib:25}: {tsize} ({tsize / total_size * 100:.2f}%)"
        byfile[k] = []
        print(k)
        for section, size in e["sections"].items():
            if size < cutoff:
                continue
            k2 = f"{section:15}: {size} ({size / tsize * 100:.0f}%)"
            byfile[k].append(k2)
            print(f"  {k2}")

    accounted_size = sum(e["total"] for e in sizes_by_lib.values())
    accounted_perc = accounted_size / total_size * 100
    module_accounted_perc = (
        sum(d["_"] for d in sizes_by_module.values()) / total_size * 100
    )
    print(
        f"Total size: {total_size} ({accounted_perc:.0f}% accounted; {module_accounted_perc:.0f}% by module)"
    )

    bymodule = {}

    def add_rec(name: str, trg: dict, d: dict):
        trg2 = {}
        n = d["_"]
        if n < cutoff:
            return
        trg[f"{name:25}: {n}"] = trg2
        for k, v in sorted(d.items(), key=lambda x: 0 if x[0] == "_" else x[1]["_"], reverse=True):
            if k == "_":
                continue
            add_rec(k, trg2, v)

    for k, v in sorted(sizes_by_module.items(), key=lambda x: x[1]["_"], reverse=True):
        add_rec(k, bymodule, v)

    with open("output.json", "w") as f:
        f.write(
            json.dumps(
                {
                    "byfile": byfile,
                    "bymodule": bymodule,
                    "total_size": total_size,
                    "accounted_size": accounted_size,
                    "accounted_perc": accounted_perc,
                    "module_accounted_perc": module_accounted_perc,
                },
                indent=2,
            )
        )

if __name__ == "__main__":
    main()
